import * as toml from "toml";
import { Resource } from "./resource";
import {
  UrlProvider,
  ModrinthProvider,
  CurseForgeProvider,
  GitHubProvider,
} from "./provider";
import { HashFormat, isValidHashFormat } from "./enums/hash-format";
import { isValidSide, type Side } from "./enums/side";
import * as semver from "semver";

export interface Metafile {
  name: string;
  filename: string;
  side: Side;
  provider: UrlProvider;
  option: Option;
}

export interface Option {
  optional: boolean;
  default?: boolean;
  description?: string;
}

/**
 * Represents a file entry in the Packwiz index.
 */
export class IndexFileEntry {
  constructor(
    readonly file: Resource,
    readonly hash: string,
    readonly hashFormat: HashFormat,
    readonly alias?: string,
    readonly metafile: boolean = false,
    readonly preserve: boolean = false,
  ) {
    this.file = file instanceof Resource ? file : new Resource(file);
    this.hash = hash;
    this.hashFormat = isValidHashFormat(hashFormat);
    this.alias = alias;
    this.metafile = metafile;
    this.preserve = preserve;
  }

  async parse(): Promise<Metafile> {
    if (this.metafile === false) {
      throw new Error(
        "Cannot parse non-metafiles! Use `file.fetchContents()` instead to get the file contents.",
      );
    }
    const fileContent = await this.file.fetchContents();
    const parsed = toml.parse(fileContent);
    const side = isValidSide(parsed.side);

    const metafile: Metafile = {
      name: parsed.name,
      filename: parsed.filename,
      side: side,
      provider: this.parseProvider(parsed),
      option: {
        optional: parsed.option?.optional || false,
        default: parsed.option?.default,
        description: parsed.option?.description,
      },
    };

    return metafile;
  }

  private parseProvider(parsed: any): UrlProvider {
    if (parsed.update?.modrinth) {
      return new ModrinthProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
        parsed.update.modrinth["mod-id"],
        parsed.update.modrinth["version"],
      );
    } else if (parsed.update?.curseforge) {
      return new CurseForgeProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        parsed.download.mode,
        parsed.update.curseforge["file-id"],
        parsed.update.curseforge["project-id"],
      );
    } else if (parsed.update?.github) {
      return new GitHubProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
        parsed.update.github.branch,
        parsed.update.github.regex,
        parsed.update.github.slug,
        parsed.update.github.tag,
      );
    } else if (parsed.download) {
      return new UrlProvider(
        parsed.download.hash,
        isValidHashFormat(parsed.download["hash-format"]),
        new Resource(parsed.download.url),
      );
    } else {
      throw new Error("Unknown provider in TOML.");
    }
  }
}

/**
 * Represents the structure of a Packwiz index file, as well as providing its location.
 */
export interface PackwizIndex {
  location: Resource;
  hash: string;
  hashFormat: HashFormat;
  files: IndexFileEntry[];
}

export interface PackwizVersions {
  minecraft: string;
  fabric?: string;
  forge?: string;
  neoforge?: string;
  quilt?: string;
  liteloader?: string;
}

export class Packwiz {
  constructor(
    readonly location: Resource,
    readonly index: PackwizIndex,
    readonly name: string,
    readonly packFormat: string = "packwiz:1.0.0",
    readonly authors: Array<string>,
    readonly description: string,
    readonly version: string,
    readonly versions: PackwizVersions,
  ) {
    const packwizJsMaxSupportedVersion = "1.1.0";
    const packFormatSemver = semver.clean(packFormat.split(":")[1]);
    if (!packFormat.startsWith("packwiz:")) {
      throw new Error("Invalid packwiz pack format!");
    } else if (!packFormatSemver) {
      throw new Error("Unable to parse pack format!");
    } else if (semver.gt(packFormatSemver, packwizJsMaxSupportedVersion)) {
      throw new Error(
        `Pack format of ${packFormat} is higher than what PackwizJS supports! Please report this upstream at https://www.coastalcommits.com/GalacticFactory/PackwizJS/issues`,
      );
    }
  }
}

/**
 * Parses a packwiz.toml file and returns its contents.
 *
 * @param filePath - The path of the TOML file.
 * @returns The parsed file as a structured class.
 */
export async function parsePackwiz(filePath: string): Promise<Packwiz> {
  const packwizFile = new Resource(filePath);
  const parsedPackwizFile = toml.parse(await packwizFile.fetchContents());

  const indexFile = packwizFile.parent.join(parsedPackwizFile.index.file);
  const parsedIndexFile = toml.parse(await indexFile.fetchContents());
  const indexHashFormat = isValidHashFormat(parsedIndexFile["hash-format"]);

  return new Packwiz(
    packwizFile,
    {
      location: indexFile,
      hash: parsedPackwizFile.index.hash,
      hashFormat: indexHashFormat,
      files: parsedIndexFile.files.map(
        (file: {
          file: string;
          hash: string;
          hashFormat: HashFormat;
          alias?: string;
          metafile?: boolean;
          preserve?: boolean;
        }) => {
          return new IndexFileEntry(
            packwizFile.parent.join(file.file),
            file.hash,
            file.hashFormat ?? indexHashFormat,
            file.alias,
            file.metafile ?? false,
            file.preserve ?? false,
          );
        },
      ),
    },
    parsedPackwizFile.name,
    parsedPackwizFile["pack-format"],
    parsedPackwizFile.author?.split(",").map((s: string) => s.trim()),
    parsedPackwizFile.description,
    parsedPackwizFile.version,
    {
      minecraft: parsedPackwizFile.versions.minecraft,
      fabric: parsedPackwizFile.versions.fabric,
      forge: parsedPackwizFile.versions.forge,
      neoforge: parsedPackwizFile.versions.neoforge,
      quilt: parsedPackwizFile.versions.quilt,
      liteloader: parsedPackwizFile.versions.liteloader,
    },
  );
}
