import * as fs from "fs/promises";
import { URL } from "url";
import * as path from "path";

export class Resource {
  constructor(public readonly path: string) { }

  toString(): string {
    return this.path;
  }

  get isUrl(): boolean {
    try {
      new URL(this.path);
      return true;
    } catch {
      return false;
    }
  }

  get name(): string {
    return this.isUrl
      ? new URL(this.path).pathname.split("/").pop() || ""
      : path.basename(this.path);
  }

  get parent(): Resource {
    if (this.isUrl) {
      const url = new URL(this.path);
      url.pathname = path.dirname(url.pathname);
      return new Resource(url.toString());
    }
    return new Resource(path.dirname(this.path));
  }

  get ext(): string {
    return this.isUrl
      ? path.extname(new URL(this.path).pathname)
      : path.extname(this.path);
  }

  async exists(): Promise<boolean> {
    if (this.isUrl) {
      const response = await fetch(this.path);
      if (!response.ok) return false;
      return true;
    } else {
      return fs.exists(this.path);
    }
  }

  async fetchContents(): Promise<string> {
    if (this.isUrl) {
      const response = await fetch(this.path);
      if (!response.ok)
        throw new Error(`Failed to fetch ${this.path}: ${response.statusText}`);
      return await response.text();
    }
    return fs.readFile(this.path, { encoding: "utf8" });
  }

  join(...segments: string[]) {
    if (this.isUrl) {
      const url = new URL(this.path);
      url.pathname = path.join(url.pathname, ...segments);
      return new Resource(url.toString());
    }
    return new Resource(path.join(this.path, ...segments));
  }
}
